# coding=utf-8

from os import path

from flask_sqlalchemy import SQLAlchemy
from flask.cli import AppGroup

from os import makedirs

db = SQLAlchemy()
db_cli = AppGroup('database')


def init_app(app):
    """Initalizes the application database (SQL).

    :param app: The Flask application object.
    """
    db.init_app(app)
    app.cli.add_command(db_cli)

    try:
        # Ensure the database folder exists
        database_folder = path.join(app.instance_path, app.config['DATABASE_FOLDER'])
        if not path.isdir(database_folder):
            makedirs(database_folder)
            raise Exception(f'Directory not found, so just created: {database_folder}')
    except OSError as ex:
        app.logger.error(str(ex))
    except Exception as ex:
        app.logger.exception(str(ex))

    # Check if sql database exists
    if not path.isfile(app.config['DATABASE_FILENAME']):
        app.logger.warning(f'Database SQL not found! File: {app.config["DATABASE_FILENAME"]}')


@db_cli.command('create')
def create_database():
    """Create the database schema."""
    db.create_all()
    print('Database SQL created')


@db_cli.command('drop')
def drop_database():
    """Drop the database."""
    response = input('Are you really sure to delete the SQL database and all the data it contains? [Y/n]:')
    if response.lower() == 'y':
        db.drop_all()
        print('Database SQL dropped')



# coding=utf-8

from flask_jwt_extended import JWTManager

jwt = JWTManager()


def init_app(app):
    """Initalizes the application auth.

    :param app: The Flask application object.
    """
    jwt.init_app(app)



# coding=utf-8

from mercury.controllers.notification import NotificationListAPI, NotificationAPI
from mercury.controllers.user import UserListAPI, UserAPI, UserLoginAPI


def init_api(api):
    """Initalizes the application Api routes.

    :param api: The Flask application's Api object.
    """
    api.add_resource(UserListAPI, f'/mercury/api/users/', endpoint='users')
    api.add_resource(UserAPI, f'/mercury/api/users/<int:id>', endpoint='user')
    api.add_resource(UserLoginAPI, f'/mercury/api/users/login/', endpoint='users_login')
    api.add_resource(NotificationListAPI, f'/mercury/api/users/<int:user_id>/notifications/',
                     endpoint='notifications')
    api.add_resource(NotificationAPI, f'/mercury/api/users/<int:user_id>/notifications/<string:_id>',
                     endpoint='notification')



# coding=utf-8

from celery import Celery
from celery.schedules import crontab

from os import path, makedirs

celery = Celery(__name__)


def init_app(app):
    """Initalizes the application with the extension.

    :param app: The Flask application object.
    """
    celery.conf.broker_url = app.config['BROKER_URL']
    celery.conf.update(app.config)
    celery.config_from_object(CeleryBeatConfig())  # Load Celery Beat instance config

    try:
        # Ensure the celery beat folder exists
        celery_beat_folder = path.join(app.instance_path, app.config['CELERY_BEAT_FOLDER'])
        if not path.isdir(celery_beat_folder):
            makedirs(celery_beat_folder)
            raise Exception(f'Directory not found, so just created: {celery_beat_folder}')
    except OSError as ex:
        app.logger.error(str(ex))
    except Exception as ex:
        app.logger.exception(str(ex))

    try:
        # Ensure the celery logs folder exists
        celery_logs_folder = path.join(app.instance_path, app.config['CELERY_LOGS_FOLDER'])
        if not path.isdir(celery_logs_folder):
            makedirs(celery_logs_folder)
            raise Exception(f'Directory not found, so just created: {celery_logs_folder}')
    except OSError as ex:
        app.logger.error(str(ex))
    except Exception as ex:
        app.logger.exception(str(ex))

    class ContextTask(celery.Task):
        def __call__(self, *args, **kwargs):
            with app.app_context():
                return self.run(*args, **kwargs)

    celery.Task = ContextTask


class CeleryBeatConfig(object):
    def __init__(self):
        """CeleryBeatConfig constructor"""
        self.CELERY_TASK_SERIALIZER = 'json'
        self.CELERY_RESULT_SERIALIZER = 'json'
        self.CELERY_ACCEPT_CONTENT = ['json']
        self.CELERY_IMPORTS = ('mercury.services.tasks.notification', )
        self.CELERY_TIMEZONE = 'UTC'
        self.CELERY_TASK_RESULT_EXPIRES = 30

        self.CELERYBEAT_SCHEDULE = {
            'route_notifications': {
                'task': 'mercury.services.tasks.notification.route_notifications',
                'schedule': crontab(minute='*'),  # Every minute
            }
        }



# coding=utf-8

from mercury.services.database_sql import db

from typing import Final
from datetime import datetime
from secrets import token_hex, compare_digest


class User(db.Model):
    _PASSWORD_STORED_MAX_LENGTH: Final = 128  # Password max bytes length: 64 bytes (512 bit) * 2 = 128 bytes (because hex stored)

    id = db.Column(db.Integer, primary_key=True)
    username = db.Column(db.String(254), index=True, unique=True, nullable=False)
    password = db.Column(db.String(_PASSWORD_STORED_MAX_LENGTH), nullable=False)
    creation_datetime = db.Column(db.DateTime, nullable=False, default=datetime.utcnow())
    active = db.Column(db.Boolean, nullable=False, default=True)
    admin = db.Column(db.Boolean, nullable=False, default=False)

    access_token = None

    def __repr__(self):
        """Returns user representation.

        :return: Username that represent user.
        """
        return '<User %r>' % self.username

    def generate_password(self):
        """Generate a password and save it to user."""
        self.password = token_hex(int(self._PASSWORD_STORED_MAX_LENGTH / 2))

    def verify_password(self, password):
        """Verify the passed password against user's password.

        :param password: Password to compare.
        :return: Outcome of the comparison, True if password is correct, otherwise False.
        """
        return compare_digest(self.password, password)



# coding=utf-8

from mercury.services import notification as services_notification
from mercury.services import user as services_user
from mercury.services.custom_exceptions import MethodVersionNotFound

from flask import abort, request
from flask_restful import Resource, marshal
from flask_jwt_extended import jwt_required, get_jwt_identity


class NotificationListAPI(Resource):
    decorators = [jwt_required]

    def __init__(self):
        """NotificationListAPI constructor."""
        self.reqparse = services_notification.get_request_parser()
        super(NotificationListAPI, self).__init__()

    def get(self, user_id):
        """GET

        :param user_id: User's id to find.
        :return: All notifications.
        """
        if request.headers.get('Accept-Version', '1.0') == '1.0':
            if get_jwt_identity() != user_id and (not services_user.check_if_user_is_admin(get_jwt_identity())):
                abort(401)
            return {'notifications': [marshal(notification, services_notification.notification_fields) for notification
                                      in services_notification.select_notifications(user_id)]}
        raise MethodVersionNotFound()

    def post(self, user_id):
        """POST

        :param user_id: User's id to find.
        :return: Persisted notification's base informations as JSON or error.
        """
        if request.headers.get('Accept-Version', '1.0') == '1.0':
            if get_jwt_identity() != user_id and (not services_user.check_if_user_is_admin(get_jwt_identity())):
                abort(401)
            if not request.json:
                abort(400)
            return {'notification': marshal(services_notification.insert_notification(request.json, user_id),
                                            services_notification.notification_fields)}, 201
        raise MethodVersionNotFound()


class NotificationAPI(Resource):
    decorators = [jwt_required]

    def __init__(self):
        """NotificationAPI constructor."""
        self.reqparse = services_notification.get_request_parser()
        super(NotificationAPI, self).__init__()

    def get(self, user_id, _id):
        """GET

        :param user_id: User's id to find.
        :param _id: Notification's id to find.
        :return: Notification found as JSON.
        """
        if request.headers.get('Accept-Version', '1.0') == '1.0':
            if get_jwt_identity() != user_id and (not services_user.check_if_user_is_admin(get_jwt_identity())):
                abort(401)
            return {'notification': marshal(services_notification.select_notification(_id, user_id),
                                            services_notification.notification_fields)}
        raise MethodVersionNotFound()

    def put(self, user_id, _id):
        """PUT

        :param user_id: User's id to find.
        :param _id: Notification's id to find.
        :return: Persisted notification's base informations as JSON or error.
        """
        if request.headers.get('Accept-Version', '1.0') == '1.0':
            if get_jwt_identity() != user_id and (not services_user.check_if_user_is_admin(get_jwt_identity())):
                abort(401)
            if not request.json:
                abort(400)
            result = services_notification.update_notification(_id, request.json, user_id)
            if result is None:
                return {'result': False}
            return {'notification': marshal(result, services_notification.notification_fields)}
        raise MethodVersionNotFound()

    def delete(self, user_id, _id):
        """DELETE

        :param user_id: User's id to find.
        :param _id: Notification's id to find.
        :return: True if elimination was successful or False if elimination was not possible.
        """
        if request.headers.get('Accept-Version', '1.0') == '1.0':
            if get_jwt_identity() != user_id and (not services_user.check_if_user_is_admin(get_jwt_identity())):
                abort(401)
            return {'result': services_notification.delete_notification(_id, user_id)}
        raise MethodVersionNotFound()



# coding=utf-8

from os import path


class Config(object):
    def __init__(self, app=None):
        """Config constructor.

        :param app: Application in which to inject the development settings.
        """
        self.app = app

        self.DEBUG = False

        # Flask
        self.SECRET_KEY = None

        # Flask-JWT-Extended
        self.JWT_SECRET_KEY = self.SECRET_KEY

        # DBs
        self.DATABASE_FOLDER = 'sqldb'
        self.DATABASE_FILENAME = path.join(self.app.instance_path, self.DATABASE_FOLDER, 'mercury.sqlite3')
        self.SQLALCHEMY_DATABASE_URI = f'sqlite:///{self.DATABASE_FILENAME}'
        self.SQLALCHEMY_TRACK_MODIFICATIONS = False
        self.MONGO_URI = None

        # Celery - for Tasks
        self.BROKER_URL = None
        self.CELERY_RESULT_BACKEND = self.MONGO_URI
        self.CELERY_LOGS_FOLDER = 'logs'
        self.CELERY_BEAT_FOLDER = 'celerybeat'

        # Email
        self.MAIL_SERVER = None
        self.MAIL_PORT = None
        self.MAIL_USE_TLS = True
        self.MAIL_USERNAME = None
        self.MAIL_PASSWORD = None
        self.MAIL_DEFAULT_SENDER = (None, None)



# coding=utf-8

from flask_pymongo import PyMongo

db = PyMongo()


def init_app(app):
    """Initalizes the application database MongoDB (NoSQL).

    :param app: The Flask application object.
    """
    db.init_app(app)



# coding=utf-8

from mercury.services import user as services_user
from mercury.services.custom_exceptions import MethodVersionNotFound

from flask import abort, request
from flask_restful import Resource, marshal
from flask_jwt_extended import jwt_required, get_jwt_identity


class UserListAPI(Resource):
    decorators = [jwt_required]

    def __init__(self):
        """UserListAPI constructor."""
        self.reqparse = services_user.get_request_parser()
        super(UserListAPI, self).__init__()

    def get(self):
        """GET

        :return: All users.
        """
        if request.headers.get('Accept-Version', '1.0') == '1.0':
            if not services_user.check_if_user_is_admin(get_jwt_identity()):
                abort(401)
            return {'users': [marshal(user, services_user.user_fields) for user in services_user.select_users()]}
        raise MethodVersionNotFound()

    def post(self):
        """POST

        :return: Persisted user's JSON or error.
        """
        if request.headers.get('Accept-Version', '1.0') == '1.0':
            if not services_user.check_if_user_is_admin(get_jwt_identity()):
                abort(401)
            if not request.json:
                abort(400)
            user = {key: value for key, value in self.reqparse.parse_args().items() if value is not None}
            return {'user': marshal(services_user.insert_user(user['username']), services_user.user_fields)}, 201
        raise MethodVersionNotFound()


class UserAPI(Resource):
    decorators = [jwt_required]

    def __init__(self):
        """UserAPI constructor."""
        self.reqparse = services_user.get_request_parser()
        super(UserAPI, self).__init__()

    def get(self, id):
        """GET

        :param id: User's id to find.
        :return: User found as JSON.
        """
        if request.headers.get('Accept-Version', '1.0') == '1.0':
            if get_jwt_identity() != id and (not services_user.check_if_user_is_admin(get_jwt_identity())):
                abort(401)
            return {'user': marshal(services_user.select_user(id), services_user.user_fields)}
        raise MethodVersionNotFound()

    def put(self, id):
        """PUT

        :param id: User's id to find.
        :return: Persisted user's base informations as JSON or error.
        """
        if request.headers.get('Accept-Version', '1.0') == '1.0':
            if get_jwt_identity() != id and (not services_user.check_if_user_is_admin(get_jwt_identity())):
                abort(401)
            if not request.json:
                abort(400)
            user = services_user.select_user(id)
            [user.__setattr__(key, value) for key, value in self.reqparse.parse_args().items() if value is not None]
            return {'user': marshal(services_user.update_user(user), services_user.user_fields)}
        raise MethodVersionNotFound()

    def delete(self, id):
        """DELETE

        :param id: User's id to find.
        :return: True if elimination was successful or False if elimination was not possible.
        """
        if request.headers.get('Accept-Version', '1.0') == '1.0':
            if get_jwt_identity() != id and (not services_user.check_if_user_is_admin(get_jwt_identity())):
                abort(401)
            return {'result': services_user.delete_user(id)}
        raise MethodVersionNotFound()


class UserLoginAPI(Resource):
    def __init__(self):
        """UserLoginAPI constructor."""
        self.reqparse = services_user.get_request_parser(is_login_request=True)
        super(UserLoginAPI, self).__init__()

    def post(self):
        """Post

        :return: User's access_token as JSON or error.
        """
        if request.headers.get('Accept-Version', '1.0') == '1.0':
            if not request.json:
                abort(400)
            user = {key: value for key, value in self.reqparse.parse_args().items() if value is not None}
            return {'user': marshal(services_user.login_user(user), services_user.user_login_fields)}
        raise MethodVersionNotFound()



# coding=utf-8

from os import path
from setuptools import setup, find_packages

version = '1.0.3'

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as file:
    long_description = file.read()

setup(
    name='mercury-py',
    packages=find_packages(include=['mercury']),
    package_data={'mercury': ['static/favicon.ico']},
    include_package_data=True,
    zip_safe=False,
    version=version,
    description='mercury-py (Mercury for Python) is a Python based microservice that allow to manage scheduled '
                'notifications sending.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords=['notifications', 'authentication', 'email', 'sms', 'push notification', 'telegram'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Communications',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3'
    ],  # https://pypi.org/classifiers/
    author='Simone Perini',
    author_email='perinisimone98@gmail.com',
    license='BSD-3-Clause',
    url='https://github.com/CoffeePerry/mercury-py',
    download_url=f'https://github.com/CoffeePerry/mercury-py/archive/{version}.tar.gz',
    install_requires=[
        'Flask==1.1.2',
        'Flask-RESTful==0.3.8',
        'Flask-SQLAlchemy==2.4.4',
        'Flask-PyMongo==2.3.0',
        'Flask-JWT-Extended==3.25.0',
        'Flask-Mail==0.9.1',
        'celery==4.4.5'
    ]
)



# coding=utf-8

from .services.database_sql import init_app as init_database_sql
from .services.database_nosql_mongo import init_app as init_database_nosql_mongo
from .services.auth import init_app as init_auth
from .services.tasks import init_app as init_tasks
from .services.tasks.notification import init_app as init_notification

from os import path, makedirs

from flask import Flask, send_from_directory
from flask_restful import Api

CONFIG_FILENAME = 'config.py'

api = Api()


def create_app():
    """Application factory.

    :return: An application instance.
    """
    # Create and configure the app
    app = Flask(__name__, instance_relative_config=True)

    # Load the default config
    from .services.config import Config
    app.config.from_object(Config(app))

    try:
        # Ensure the instance folder exists
        if not path.isdir(app.instance_path):
            makedirs(app.instance_path)
            raise Exception(f'Directory not found, so just created: {app.instance_path}.'
                            f'Put file "{CONFIG_FILENAME}" inside, please.')
    except OSError as ex:
        app.logger.error(str(ex))
    except Exception as ex:
        app.logger.exception(str(ex))

    # Load the instance config
    app.config.from_pyfile(path.join(app.instance_path, CONFIG_FILENAME))

    secret_key = app.config.get('SECRET_KEY')
    if secret_key is None:
        app.logger.error(f'Set variable SECRET_KEY with random string in file: '
                         f'{path.join(app.instance_path, CONFIG_FILENAME)}')

    init_database_sql(app)
    init_database_nosql_mongo(app)

    init_auth(app)

    # Load routes after load other services
    from .services.routes import init_api
    init_api(api)
    api.init_app(app)

    init_tasks(app)
    init_notification(app)

    # Base page for check if service is ready
    @app.route(f'/mercury/api/')
    def index():
        """Return Mercury index page.

        :return: Mercury index page.
        """
        return f'Mercury online!'

    @app.route('/favicon.ico')
    def favicon():
        """Return Mercury favicon.ico.

        :return: Mercury favicon.ico.
        """
        return send_from_directory(path.join(app.root_path, 'static'),
                                   'favicon.ico', mimetype='image/vnd.microsoft.icon')

    @app.cli.command('info')
    def info():
        """Print Mercury info.

        :return: Mercury info.
        """
        print(f'Mercury CLI ready!')

    return app



# coding=utf-8

from .. import celery
from mercury.services.notification import find_notifications_to_dispatch, update_notification

from datetime import datetime
from base64 import b64decode
from os import linesep

from flask_mail import Mail, Message

mail = Mail()


def init_app(app):
    """Initalizes the application with the extension.

    :param app: The Flask application object.
    """
    mail.init_app(app)


@celery.task()
def route_notifications():
    """Celery task routine to find and route notifications to dispatch to correct notification channel
    (executed by celery worker).

    :return: Operations logs.
    """
    notifications = find_notifications_to_dispatch()
    if notifications is None or notifications.count() < 1:
        return 'Nothing to dispatch.'
    result = ''
    for notification in notifications:
        result = f'{result}{linesep}{route_notification(notification)}'
    return result


def route_notification(notification):
    """Routes a notification to correct notification channel.

    :param notification: Notification to route to correct notification channel.
    :return: Operations logs.
    """
    return {
        'email': route_notification_mail(notification)
    }.get(notification['category'], 'email')


def route_notification_mail(notification):
    """Routes a notification to mail notification channel.

    :param notification: Notification to route to mail notification channel.
    :return: Operations logs.
    """
    recipients = notification['recipients']
    if isinstance(recipients, str):
        recipients = [recipients]
    msg = Message(recipients=recipients, subject=notification['subject'],
                  body=notification.get('body'), html=notification.get('html'),
                  cc=notification.get('cc'), bcc=notification.get('bcc'), reply_to=notification.get('reply_to'))
    try:
        attachments = notification.get('attachments')
        if attachments is not None:
            [msg.attach(filename=attachment.get('filename'), content_type=attachment.get('content_type'),
                        data=b64decode(attachment['data'])) for attachment in attachments]
    except Exception as ex:
        raise Exception(f'Mail attachments with _id {notification["_id"]} failed to load at '
                        f'{datetime.now().strftime("%Y-%m-%d %H:%M:%S")}!{linesep}Detail: {str(ex)}')
    try:
        mail.send(msg)
    except Exception as ex:
        raise Exception(f'Dispatch mail with _id {notification["_id"]} failed at '
                        f'{datetime.now().strftime("%Y-%m-%d %H:%M:%S")}!{linesep}Detail: {str(ex)}')
    success_msg = f'Successfully dispatched mail with _id {notification["_id"]} at ' \
                  f'{datetime.now().strftime("%Y-%m-%d %H:%M:%S")}.'
    try:
        notification['datetime_dispatch'] = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        result = update_notification(notification['_id'], notification)
        if result is None:
            raise Exception('Acknowledged: False.')
    except Exception as ex:
        raise Exception(f'Update mail datetime_dispatch with _id {notification["_id"]} failed at '
                        f'{datetime.now().strftime("%Y-%m-%d %H:%M:%S")}!{linesep}Detail: {str(ex)}')
    return f'{success_msg}{linesep}Successfully updated mail datetime_dispatch with _id {notification["_id"]}.'



# coding=utf-8

from mercury.models.user import User
from mercury.services.database_sql import db, db_cli

from datetime import timedelta

from werkzeug.exceptions import BadRequest, Unauthorized, Forbidden
from flask_restful import fields, reqparse, inputs
from flask_jwt_extended import create_access_token
from click import argument

"""Fields to marshal user to JSON."""
user_fields = {
    'username': fields.String,
    'creation_datetime': fields.String,
    'active': fields.Boolean,
    'links': {
        'self': fields.Url('user')  # Replaces User ID with User Uri (HATEOAS) through endpoint 'user'
    }
}

"""Fields to marshal user login info to JSON."""
user_login_fields = dict(user_fields)
user_login_fields['access_token'] = fields.String


def get_request_parser(request_parser=None, is_login_request=False):
    """Get request parser for user.

    :param request_parser: If exists, add request parser argument to request_parser param.
    :param is_login_request: If True, add request parser argument password to request_parser param.
    :return: User request parser.
    """
    if request_parser is None:
        result = reqparse.RequestParser()
    else:
        result = request_parser
    result.add_argument('username', type=str, required=True, help='No user username provided', location='json')
    if is_login_request:
        result.add_argument('password', type=str, required=True, help='No user password provided', location='json')
    result.add_argument('active', type=inputs.boolean, required=False, help='No user active provided', location='json')
    return result


'''CRUD Functions'''


def select_user(id):
    """Get user by id param.

    :param id: User's id to find.
    :return: User found.
    """
    return User.query.get_or_404(id)


def select_users():
    """Get all users.

    :return: All users.
    """
    return User.query.all()


def insert_user(username):
    """Post new user from username param.

    :param username: User's username to persist.
    :return: Persisted user or error.
    """
    user = User()
    user.username = username
    user.generate_password()
    db.session.add(user)
    db.session.commit()
    return user


def update_user(user):
    """Put the user passed by user param.

    :param user: User to persist.
    :return: Persisted user or error.
    """
    if user is None:
        return None
    db.session.commit()
    return user


def delete_user(id):
    """Delete the user that have the passed user id.

    :param id: User's id to find.
    :return: True if elimination was successful or False if elimination was not possible.
    """
    user = User.query.get_or_404(id)
    db.session.delete(user)
    db.session.commit()
    return True


'''Other Functions'''


def login_user(user):
    """Login user from user param.

    :param user: User to login.
    :return: User's access_token or error.
    """
    username = user.get('username')
    if username is None:
        raise BadRequest('Missing username parameter')
    password = user.get('password')
    if password is None:
        raise BadRequest('Missing password parameter')
    db_user = User.query.filter(User.username == username).scalar()
    if db_user and db_user.verify_password(password):
        if not db_user.active:
            raise Forbidden('User inactive')
        db_user.access_token = create_access_token(identity=db_user.id, expires_delta=timedelta(days=1))
        return db_user
    else:
        raise Unauthorized('Bad username or password')


def check_if_user_is_admin(id):
    """Check if user by id param is admin.

    :param id: User's id to find.
    :return: Outcome of the comparison, True if user is admin, otherwise False.
    """
    user = User.query.get(id)
    if user is None:
        return False
    return user.admin


@db_cli.command('register')
@argument('username')
def register_user(username):
    """Register new user.

    :param username: New user's username.
    """
    if not username:
        print('Please, enter an unique Username.')
    else:
        response = input(f'Are you really sure to register new user {username}? [Y/n]:')
        if response.lower() == 'y':
            print('Creation in progress...')
            user = insert_user(username)
            print(f'{user} successfully created')
        else:
            print('Aborted.')



# coding=utf-8

from mercury.services.database_nosql_mongo import db as mongo

from datetime import datetime
from bson import ObjectId

from flask_restful import fields, reqparse

"""Fields to marshal notification to JSON."""
notification_fields = {
    'category': fields.String,
    'datetime_schedule': fields.String,
    'datetime_dispatch': fields.String,
    'links': {
        # Replaces Notification ID with Notification Uri (HATEOAS) through endpoint 'notification'
        'self': fields.Url('notification')
    }
}


'''CRUD Functions'''


def get_request_parser(request_parser=None):
    """Get request parser for notification.

    :param request_parser: If exists, add request parser argument to request_parser param.
    :return: Notification request parser.
    """
    if request_parser is None:
        result = reqparse.RequestParser()
    else:
        result = request_parser
    result.add_argument('category', type=str, required=True, help='No notification category provided', location='json')
    result.add_argument('datetime_schedule', type=lambda x: datetime.strptime(x, '%Y-%m-%d %H:%M:%S'), required=False,
                        help='No notification datetime_schedule provided', location='json')
    return result


def select_notification(id, user_id):
    """Get notification by id param and user_id.

    :param id: Notification's id to find.
    :param user_id: Notification's user_id to find.
    :return: Notification found or error 404.
    """
    return mongo.db.notification.find_one_or_404({'_id': ObjectId(id), 'user_id': user_id})


def select_notifications(user_id):
    """Get all notifications by user_id.

    :param user_id: Notification's user_id to find.
    :return: All notifications.
    """
    return mongo.db.notification.find({'user_id': user_id})


def insert_notification(notification, user_id):
    """Post new notification from notification param (MongoDB has limit of 16 megabytes per document) for user_id.

    :param notification: Notification to persist.
    :param user_id: Notification's user_id to persist.
    :return: Persisted notification's base informations or error.
    """
    notification['user_id'] = user_id
    notification['datetime_schedule'] = notification.get('datetime_schedule',
                                                         datetime.now().strftime('%Y-%m-%d %H:%M:%S'))
    notification.pop('_id', None)
    return {
        '_id': mongo.db.notification.insert_one(notification).inserted_id,
        'user_id': user_id,
        'category': notification['category'],
        'datetime_schedule': notification['datetime_schedule']
    }


def update_notification(id, notification, user_id=None):
    """Put the notification passed by notification param (MongoDB has limit of 16 megabytes per document) for user_id.

    :param id: Notification's id to find.
    :param user_id: Notification's user_id to find.
    :param notification: Notification to persist.
    :return: Persisted notification's base informations or error.
    """
    if user_id is None:
        user_id = notification['user_id']
    else:
        notification['user_id'] = user_id
    notification.pop('_id', None)
    notification_found = mongo.db.notification.find_one_or_404({'_id': ObjectId(id), 'user_id': user_id})
    if mongo.db.notification.update_one({'_id': notification_found['_id']}, {'$set': notification}).acknowledged:
        return {
            '_id': id,
            'user_id': user_id,
            'category': notification['category'],
            'datetime_schedule': notification.get('datetime_schedule', notification_found['datetime_schedule'])
        }
    else:
        return None


def delete_notification(id, user_id):
    """Delete the notification that have the passed notification id.

    :param id: Notification's id to find.
    :param user_id: Notification's user_id to find.
    :return: True if elimination was successful or False if elimination was not possible.
    """
    return mongo.db.notification.remove({'_id': ObjectId(id), 'user_id': user_id})['ok'] == 1.0


'''Other Functions'''


def find_notifications_to_dispatch():
    """Find all notifications that have datetime_schedule lower than now (local time) and that they haven't already been
     dispatched.

    :return: Cursor to manage notifications to dispatch.
    """
    return mongo.db.notification.find({
        'datetime_schedule': {'$lte': datetime.now().strftime('%Y-%m-%d %H:%M:%S')},
        'datetime_dispatch': None
    })
